// asmsh.c - copyright 1998-2007 Bruce Tomlin

#define versionName "SH family assembler"
#include "asmx.h"

enum
{
    CPU_SH1,
    CPU_SH2,
    CPU_SH3,
    CPU_SH4
};

#define LIMIT 0xFFF

enum
{
    o_Inherent,     // implied instructions
    o_Branch,       // relative branch instructions
    o_Imm8,        // 8-bit immediate
    o_MOV,
    o_OneReg,
    o_TwoReg,
    o_ShortLog,
    o_R0Shift,
    o_Ref,
    o_ADD,
    o_AND,
    o_CMP,
    o_OR,
    o_TST,
    o_XOR,
    o_MAC,
    o_LDC,
    o_LDS,
    o_STC,
    o_STS,

//  o_Foo = o_LabelOp,
};

enum
{
    WID_B, WID_W, WID_L // byte, word, or long
};

const char SH_GPR[] = "R0 R1 R2 R3 R4 R5 R6 R7 R8 R9 R10 R11 R12 R13 R14 R15 SP";
const char CR[] = "SR GBR VBR";
const char SR[] = "MACH MACL PR PC";
const flags[] = {0x3000, 0x3002, 0x3003, 0x3006, 0x3007, 0x4015, 0x4011, 0x200C};


struct OpcdRec SH_opcdTab[] =
{
    {"CLRT",    o_Inherent, 0x8},
    {"CLRMAC",  o_Inherent, 0x28},
    {"DIV0U",   o_Inherent, 0x19},
    {"NOP",     o_Inherent, 0x9},
    {"RTE",     o_Inherent, 0x2B},
    {"RTS",     o_Inherent, 0xB},
    {"SETT",    o_Inherent, 0x18},
    {"SLEEP",   o_Inherent, 0x1B},

    {"BF",      o_Branch,   0x8B00},
    {"BT",      o_Branch,   0x8900},
    {"BRA",     o_Branch,   0xA000},
    {"BSR",     o_Branch,   0xB000},

    {"TRAPA",   o_Imm8,     0xC300},

    {"MOV",     o_MOV,      0xE000},
    {"MOV.B",   o_MOV,      0x0000 + WID_B},
    {"MOV.W",   o_MOV,      0x0000 + WID_W},
    {"MOV.L",   o_MOV,      0x0000 + WID_L},
    
    {"MOVA",    o_R0Shift,  0xC700},
    
    {"BRAF",    o_OneReg,   0x23},
    {"BSRF",    o_OneReg,   0x3},
    {"DT",      o_OneReg,   0x4010},
    {"MOVT",    o_OneReg,   0x0029},
    {"ROTL",    o_OneReg,   0x4004},
    {"ROTR",    o_OneReg,   0x4005},
    {"ROTCL",   o_OneReg,   0x4024},
    {"ROTCR",   o_OneReg,   0x4025},
    {"SHAL",    o_OneReg,   0x4020},
    {"SHAR",    o_OneReg,   0x4021},
    {"SHLL",    o_OneReg,   0x4000},
    {"SHLR",    o_OneReg,   0x4001},
    {"SHLL2",   o_OneReg,   0x4008},
    {"SHLR2",   o_OneReg,   0x4009},
    {"SHLL8",   o_OneReg,   0x4018},
    {"SHLR8",   o_OneReg,   0x4019},
    {"SHLL16",  o_OneReg,   0x4028},
    {"SHLR16",  o_OneReg,   0x4029},

    {"ADDC",    o_TwoReg,   0x300E},
    {"ADDV",    o_TwoReg,   0x300F},
    {"DIV1",    o_TwoReg,   0x3004},
    {"DIV0S",   o_TwoReg,   0x2007},
    {"DMULS.L", o_TwoReg,   0x300D},
    {"DMULU.L", o_TwoReg,   0x3005},
    {"EXTS.B",  o_TwoReg,   0x600E + WID_B},
    {"EXTS.W",  o_TwoReg,   0x600E + WID_W},
    {"EXTU.B",  o_TwoReg,   0x600C + WID_B},
    {"EXTU.W",  o_TwoReg,   0x600C + WID_W},
    {"MUL.L",   o_TwoReg,   0x0007},
    {"MULS.W",  o_TwoReg,   0x200F},
    {"MULU.W",  o_TwoReg,   0x200E},
    {"NEG",     o_TwoReg,   0x600B},
    {"NEGC",    o_TwoReg,   0x600A},
    {"NOT",     o_TwoReg,   0x6007},
    {"SUB",     o_TwoReg,   0x3008},
    {"SUBC",    o_TwoReg,   0x300A},
    {"SUBV",    o_TwoReg,   0x300B},
    {"SWAP.B",  o_TwoReg,   0x6008 + WID_B},
    {"SWAP.W",  o_TwoReg,   0x6008 + WID_W},
    {"XTRCT",   o_TwoReg,   0x200D},
    
    {"AND.B",   o_ShortLog, 0xCD00},
    {"OR.B",    o_ShortLog, 0xCF00},
    {"TST.B",   o_ShortLog, 0xCC00},
    {"XOR.B",   o_ShortLog, 0xCE00},

    {"ADD",     o_ADD,      0x3000},
    
    {"AND",     o_AND,      0x0000},
    {"CMP",     o_CMP,      0x0000},
    {"OR",      o_OR,       0x0000},
    {"TST",     o_TST,      0x0000},
    {"XOR",     o_XOR,      0x0000},
    
    {"MAC.L",   o_MAC,      0x000F},
    {"MAC.W",   o_MAC,      0x400F},
    
    {"LDC",     o_LDC,      0x400E},
    {"LDC.L",   o_LDC,      0x4007},
    {"LDS",     o_LDS,      0x400A},
    {"LDS.L",   o_LDS,      0x4006},
    
    {"STC",     o_STC,      0x0002},
    {"STC.L",   o_STC,      0x4003},
    {"STS",     o_STS,      0x000A},
    {"STS.L",   o_STS,      0x4002},
    
    {"JMP",     o_Ref,      0x402B},
    {"JSR",     o_Ref,      0x400B},
    {"TAS.B",   o_Ref,      0x401B},

    {"",        o_Illegal,  0}
};


// --------------------------------------------------------------

int EvalShift(short size)
{
    int val;

    val = Eval();
    val = val - locPtr - 4;
    if (!errFlag && (((size==1) && (val & 1)) || ((size==2) && (val & 3)) || val < ~LIMIT || val >= LIMIT))
        Error("Long branch out of range");

    val = val >> size;
    return val;
}

int SH_DoCPUOpcode(int typ, int parm)
{
    int     val;
    Str255  word;
    char    *oldLine;
    int     reg1,reg2;

    switch(typ)
    {
        case o_Inherent:
            InstrW(parm);
            break;

        case o_Branch:
            oldLine = linePtr;
            if (GetWord(word) == '/')
            {
             if(curCPU<CPU_SH2||(parm!=0x8B00&&parm!=0x8900)) break;
             if (Expect("S")) break;
             parm=parm|0x400;
            }
            else linePtr = oldLine;
            
            val = EvalShift(1);
            
            if (!errFlag&&(parm!=0xA000&&parm!=0xB000)&&(val < -128 || val > 127)) Error("Offset out of range");
            
            InstrW(parm|val);
            break;

        case o_Imm8:
            if (Expect("#")) break;
            val = Eval();
            
            if (!errFlag&&(val < -256 || val > 255)) Error("Out of range!");
            
            InstrW(parm | val);
            break;

        case o_MOV:
            oldLine = linePtr;
            if (GetWord(word) == '#')
            {
             if (parm!=0xE000) break;
             val = Eval();
             
             if (Comma()) break;
             
             reg1 = GetReg(SH_GPR);
             if (reg1 == 16) reg1 = 15;
             if (CheckReg(reg1)) break;
            
             InstrW(parm | (reg1 << 8) | (val & 0xFF));
            }
            else
            {
             linePtr = oldLine;
             if (GetWord(word) == '@')
             {
              if(parm==0xE000) break; //postfixes is required
              
              oldLine = linePtr;
              if (GetWord(word) == '(')
              {
               oldLine = linePtr;
               reg1 = GetReg(SH_GPR);
               if(reg1>=0) //MOV.B @(R0,Rm),Rn
               {
                if(reg1!=0) break;
             
                if (Comma()) break;
             
                reg1 = GetReg(SH_GPR);
                if (reg1 == 16) reg1 = 15;
                if (CheckReg(reg1)) break;
                
                if (GetWord(word) != ')') break;
             
                if (Comma()) break;
             
                reg2 = GetReg(SH_GPR);
                if (reg2 == 16) reg2 = 15;
                if (CheckReg(reg2)) break;
            
                parm=parm|0xC;
                InstrW(parm | (reg2 << 8) | (reg1 << 4));
               }
               else //MOV.x @(disp,Reg),Reg
               {
                linePtr = oldLine;
                val = Eval();
                if (!errFlag && (((parm==1) && (val & 1)) || ((parm==2) && (val & 3)))) Error("Offset out of range");
                val=val>>parm;
             
                if (Comma()) break;
                  
                oldLine = linePtr;
                reg1 = GetReg(CR);
                if(reg1>=0) //MOV.x @(disp,GBR),R0
                {
                 if(reg1!=1) break;
                 
                 if (!errFlag && (val < -256 || val > 255)) Error("Offset out of range");
                
                 if (GetWord(word) != ')') break;
             
                 if (Comma()) break;
             
                 reg2 = GetReg(SH_GPR);
                 if (CheckReg(reg2)) break;
                 if (reg2 != 0) break;
            
                 parm=(parm<<8)|0xC400;
                 InstrW(parm | val);
                }
                else
                {
                 linePtr = oldLine;
                 reg1 = GetReg(SH_GPR);
                 if (reg1 == 16) reg1 = 15;
                 if (CheckReg(reg1)) break;
                 
                 if (!errFlag && (val < -16 || val > 15)) Error("Offset out of range");
                
                 if (GetWord(word) != ')') break;
             
                 if (Comma()) break;
             
                 reg2 = GetReg(SH_GPR);
                 if (reg2 == 16) reg2 = 15;
                 if (CheckReg(reg2)) break;
                 
                 if (parm<2)
                 {
                  if (reg2 != 0) break;
                  
                  parm=(parm<<8)|0x8400;
                  InstrW(parm | (reg1 << 4) | val);
                 }
                 else
                 {
                  parm=0x5000;
                  InstrW(parm | (reg2 << 8) | (reg1 << 4) | val);
                 }
                }
               }
              }
              else //MOV.x @Rm(+),Rn
              {
               linePtr = oldLine;
               reg1 = GetReg(SH_GPR);
               if (reg1 == 16) reg1 = 15;
               if (CheckReg(reg1)) break;
               
               oldLine = linePtr;
               if (GetWord(word) == '+') parm=parm|4;
               else linePtr = oldLine;
             
               if (Comma()) break;
             
               reg2 = GetReg(SH_GPR);
               if (reg2 == 16) reg2 = 15;
               if (CheckReg(reg2)) break;
            
               parm=parm|0x6000;
               InstrW(parm | (reg2 << 8) | (reg1 << 4));
              }
             }
             else
             {
              linePtr = oldLine;
              reg1 = GetReg(SH_GPR);
              if (reg1>=0)
              {
               if (reg1 == 16) reg1 = 15;
               if (CheckReg(reg1)) break;
             
               if (Comma()) break;
               
               oldLine = linePtr;
               if (GetWord(word) == '@')
               {
                if(parm==0xE000) break; //postfixes is required
                
                oldLine = linePtr;
                if (GetWord(word) == '(')
                {
                 oldLine = linePtr;
                 reg2 = GetReg(SH_GPR);
                 if (reg2>=0) //MOV.x Rm,@(R0,Rn)
                 {
                  if (reg2!=0) break;
             
                  if (Comma()) break;
             
                  reg2 = GetReg(SH_GPR);
                  if (reg2 == 16) reg2 = 15;
                  if (CheckReg(reg2)) break;
                  
                  if (GetWord(word) != ')') break;
            
                  parm=parm|4;
                  InstrW(parm | (reg2 << 8) | (reg1 << 4));
                 }
                 else
                 {
                  linePtr = oldLine;
                  val = Eval();
                  if (!errFlag && (((parm==1) && (val & 1)) || ((parm==2) && (val & 3)))) Error("Offset out of range");
                  val=val>>parm;
             
                  if (Comma()) break;
                  
                  oldLine = linePtr;
                  reg2 = GetReg(CR);
                  if (reg2>=0)
                  {
                   if (!errFlag && (val < -256 || val > 255)) Error("Offset out of range");
                  
                   if (reg2!=1) break; //GBR only allowed
                  
                   if (GetWord(word) != ')') break;
            
                   parm=(parm<<8)|0xC000;
                   InstrW(parm | val);
                  }
                  else //MOV.x Reg,@(disp,Rn)
                  {
                   if (!errFlag && (val < -16 || val > 15)) Error("Offset out of range");
                   
                   if ((parm<2)&&(reg1!=0)) break; // .b and .w can work only with R0
                   
                   linePtr = oldLine;
             
                   reg2 = GetReg(SH_GPR);
                   if (reg2 == 16) reg2 = 15;
                   if (CheckReg(reg2)) break;
                  
                   if (GetWord(word) != ')') break;
            
                   if (parm<2)
                   {
                    parm=(parm<<8)|0x8000;
                    InstrW(parm | (reg2 << 4) | val);
                   }
                   else
                   {
                    parm=0x1000;
                    InstrW(parm | (reg2 << 8) | (reg1 << 4) | val);
                   }
                  }
                 }
                }
                else
                {
                 linePtr = oldLine;
                 if (GetWord(word) == '-') parm=parm|4;
                 else linePtr = oldLine;
                 
                 reg2 = GetReg(SH_GPR);
                 if (reg2 == 16) reg2 = 15;
                 if (CheckReg(reg2)) break;
            
                 parm=parm|0x2000;
                 InstrW(parm | (reg2 << 8) | (reg1 << 4));
                }
               }
               else //MOV Rm,Rn
               {
                linePtr = oldLine;
                reg2 = GetReg(SH_GPR);
                if (reg2 == 16) reg2 = 15;
                if (CheckReg(reg2)) break;
                
                if (parm==0xE000) parm=0x6003; //postfixes not allowed
                else break;
                
                InstrW(parm | (reg2 << 8) | (reg1 << 4));
               }
              }
              else //MOV.(W/L) @(disp,PC),Rn
              {
               if ((parm==0)||(parm==0xE000)) break; //only .w and .l postfixes is allowed
               linePtr = oldLine;
               val = EvalShift(parm);
             
               if (Comma()) break;
             
               reg1 = GetReg(SH_GPR);
               if (reg1 == 16) reg1 = 15;
               if (CheckReg(reg1)) break;
            
               parm=((parm-1)<<14)|0x9000;
               InstrW(parm | (reg1 << 8) | val);
              }
             }
            }
            break;

        case o_OneReg:
            if(curCPU<CPU_SH2&&(parm==0x4010||parm==0x2003||parm==3)) break;
             
            reg1 = GetReg(SH_GPR);
            if (reg1 == 16) reg1 = 15;
            if (CheckReg(reg1)) break;
            
            InstrW(parm | (reg1 << 8));
            break;
            
        case o_TwoReg:
            if(curCPU<CPU_SH2&&(parm==0x300D||parm==0x3005||parm==7)) break;
        
            reg1 = GetReg(SH_GPR);
            if (reg1 == 16) reg1 = 15;
            if (CheckReg(reg1)) break;
            
            if (Comma()) break;
            
            reg2 = GetReg(SH_GPR);
            if (reg2 == 16) reg2 = 15;
            if (CheckReg(reg2)) break;
            
            InstrW(parm | (reg2 << 8) | (reg1 << 4));
            break;
            
        case o_R0Shift:
            val = EvalShift(2);
            
            if (Comma()) break;
            
            reg1 = GetReg(SH_GPR);
            if (reg1 != 0) break;
            
            InstrW(parm | val);
            break;
            
        case o_ShortLog:
            if (GetWord(word) != '#') break;
            val = Eval();
            if (val < -256 || val > 255) break;
            
            if (Comma()) break;
            
            if (GetWord(word) != '@') break;
            if (GetWord(word) != '(') break;
            reg1 = GetReg(SH_GPR);
            if(reg1!=0) break;
            
            if (Comma()) break;
            
            reg1 = GetReg(CR);
            if(reg1!=1) break;
            
            if (GetWord(word) != ')') break;
            
            InstrW(parm | val);
            break;

        case o_ADD:  
            oldLine = linePtr;
            if (GetWord(word) == '#')
            {
             val = Eval();
             if (val < -256 || val > 255) break;
             
             if (Comma()) break;
                              
             reg1 = GetReg(SH_GPR);
             if (reg1 == 16) reg1 = 15;
             if (CheckReg(reg1)) break;
             
             parm=parm|0x4000;
             InstrW(parm | (reg1 << 8) | val);
            }
            else
            {
             linePtr = oldLine;
             reg1 = GetReg(SH_GPR);
             if (reg1 == 16) reg1 = 15;
             if (CheckReg(reg1)) break;
             
             if (Comma()) break;
                              
             reg2 = GetReg(SH_GPR);
             if (reg2 == 16) reg2 = 15;
             if (CheckReg(reg2)) break;
             
             parm=parm|0xC;
             InstrW(parm | (reg2 << 8) | (reg1 << 4));
            }
            break;

        case o_AND:
            oldLine = linePtr;
            if (GetWord(word) == '#')
            {
             parm=parm|0xC900;
             val = Eval();
             if (val < -256 || val > 255) break;
             
             if (Comma()) break;
                              
             reg1 = GetReg(SH_GPR);
             if (reg1!=0) break;
            
             InstrW(parm | val);
            }
            else
            {
             parm=parm|0x2009;
             
             linePtr = oldLine;
             reg1 = GetReg(SH_GPR);
             if (reg1 == 16) reg1 = 15;
             if (CheckReg(reg1)) break;
             
             if (Comma()) break;
                              
             reg2 = GetReg(SH_GPR);
             if (reg2 == 16) reg2 = 15;
             if (CheckReg(reg2)) break;
            
             InstrW(parm | (reg2 << 8) | (reg1 << 4));
            }
            break;
            
        case o_CMP:
            if (GetWord(word) == '/')
            {
             reg2 = GetReg("EQ HS GE HI GT PL PZ STR");
             if (CheckReg(reg2)) break;
             parm=parm|flags[reg2];
             
             oldLine = linePtr;
             if (GetWord(word) == '#')
             {
              if(reg2!=0) break;
              else
              {
               val = Eval();
               if (val < -256 || val > 255) break;
               
               if (Comma()) break;
            
               reg1 = GetReg(SH_GPR);
               if(reg2!=0) break;
               
               parm=0x8800;
               InstrW(parm | val);
              }
             }
             else
             {
              linePtr = oldLine;
            
              reg1 = GetReg(SH_GPR);
              if (reg1 == 16) reg1 = 15;
              if (CheckReg(reg1)) break;
             
              if ((reg2<5)||(reg2==7))
              {
               if (Comma()) break;
            
               reg2 = GetReg(SH_GPR);
               if (reg2 == 16) reg2 = 15;
               if (CheckReg(reg2)) break;
            
               InstrW(parm | (reg2 << 8) | (reg1 << 4));
              }
              else InstrW(parm | (reg1 << 8));
             }
            }
            break;

        case o_OR:
            oldLine = linePtr;
            if (GetWord(word) == '#')
            {
             parm=parm|0xCB00;
             val = Eval();
             if (val < -256 || val > 255) break;
             
             if (Comma()) break;
                              
             reg1 = GetReg(SH_GPR);
             if (reg1!=0) break;
            
             InstrW(parm | val);
            }
            else
            {
             parm=parm|0x200B;
             
             linePtr = oldLine;
             reg1 = GetReg(SH_GPR);
             if (reg1 == 16) reg1 = 15;
             if (CheckReg(reg1)) break;
             
             if (Comma()) break;
                              
             reg2 = GetReg(SH_GPR);
             if (reg2 == 16) reg2 = 15;
             if (CheckReg(reg2)) break;
            
             InstrW(parm | (reg2 << 8) | (reg1 << 4));
            }
            break;

        case o_TST:
            oldLine = linePtr;
            if (GetWord(word) == '#')
            {
             parm=parm|0xC800;
             val = Eval();
             
             if (Comma()) break;
                              
             reg1 = GetReg(SH_GPR);
             if (reg1 != 0) break;
            
             InstrW(parm | (val & 0xFF));
            }
            else
            {
             parm=parm|0x2008;
             
             linePtr = oldLine;
             reg1 = GetReg(SH_GPR);
             if (reg1 == 16) reg1 = 15;
             if (CheckReg(reg1)) break;
             
             if (Comma()) break;
                              
             reg2 = GetReg(SH_GPR);
             if (reg2 == 16) reg2 = 15;
             if (CheckReg(reg2)) break;
            
             InstrW(parm | (reg2 << 8) | (reg1 << 4));
            }
            break;

        case o_XOR:
            oldLine = linePtr;
            if (GetWord(word) == '#')
            {
             parm=parm|0xCA00;
             val = Eval();
             
             if (Comma()) break;
                              
             reg1 = GetReg(SH_GPR);
             if (reg1 != 0) break;
            
             InstrW(parm | (val & 0xFF));
            }
            else
            {
             parm=parm|0x200A;
             
             linePtr = oldLine;
             reg1 = GetReg(SH_GPR);
             if (reg1 == 16) reg1 = 15;
             if (CheckReg(reg1)) break;
             
             if (Comma()) break;
                              
             reg2 = GetReg(SH_GPR);
             if (reg2 == 16) reg2 = 15;
             if (CheckReg(reg2)) break;
            
             InstrW(parm | (reg2 << 8) | (reg1 << 4));
            }
            break;
            
        case o_Ref:
            if (GetWord(word) == '@')
            {
             reg1 = GetReg(SH_GPR);
             if (reg1 == 16) reg1 = 15;
             if (CheckReg(reg1)) break;
            
             InstrW(parm | (reg1 << 8));
            }
            break;
            
        case o_MAC:
            if(curCPU<CPU_SH2&&parm==0xF) break;
            
            if (GetWord(word) != '@') break;
            reg1 = GetReg(SH_GPR);
            if (reg1 == 16) reg1 = 15;
            if (CheckReg(reg1)) break;
            if (GetWord(word) != '+') break;
            
            if (Comma()) break;
            
            if (GetWord(word) != '@') break;
            reg2 = GetReg(SH_GPR);
            if (reg2 == 16) reg2 = 15;
            if (CheckReg(reg2)) break;
            if (GetWord(word) != '+') break;
            
            InstrW(parm | (reg2 << 8) | (reg1 << 4));
            break;
            
        case o_LDC:
            if (parm==0x4007)
            {
             if (GetWord(word) != '@') break;
            }
             
            reg1 = GetReg(SH_GPR);
            if (reg1 == 16) reg1 = 15;
            if (CheckReg(reg1)) break;
            
            if (parm==0x4007)
            {
             if (GetWord(word) != '+') break;
            }
            
            if (Comma()) break;
            
            reg2 = GetReg(CR);
            if (CheckReg(reg2)) break;
            
            InstrW(parm | (reg1 << 8) | (reg2 << 4));
            break;
            
        case o_LDS:
            if (parm==0x4006)
            {
             if (GetWord(word) != '@') break;
            }
             
            reg1 = GetReg(SH_GPR);
            if (reg1 == 16) reg1 = 15;
            if (CheckReg(reg1)) break;
            
            if (parm==0x4006)
            { 
             if (GetWord(word) != '+') break;
            }
            
            if (Comma()) break;
            
            reg2 = GetReg(SR);
            if (CheckReg(reg2)) break;
            if (reg2>2) break;
            
            InstrW(parm | (reg1 << 8) | (reg2 << 4));
            break;
            
        case o_STC:
            reg1 = GetReg(CR);
            if (CheckReg(reg1)) break;
            
            if (Comma()) break;
            
            if (parm==0x4003)
            {
             if (GetWord(word) != '@') break;
             if (GetWord(word) != '-') break;
            }
             
            reg2 = GetReg(SH_GPR);
            if (reg2 == 16) reg2 = 15;
            if (CheckReg(reg2)) break;
            
            InstrW(parm | (reg2 << 8) | (reg1 << 4));
            break;
            
        case o_STS:
            reg1 = GetReg(SR);
            if (CheckReg(reg1)) break;
            if (reg1>2) break;
            
            if (Comma()) break;
            
            if (parm==0x4002)
            {
             if (GetWord(word) != '@') break;
             if (GetWord(word) != '-') break;
            }
            
            reg2 = GetReg(SH_GPR);
            if (reg2 == 16) reg2 = 15;
            if (CheckReg(reg2)) break;
            
            InstrW(parm | (reg2 << 8) | (reg1 << 4));
            break;

        default:
            return 0;
            break;
    }

    if (locPtr & 1)
    {
        Error("Code at non-word-aligned address");
        // deposit an extra byte to reset alignment and prevent further errors
        InstrAddB(0);
        // note: Inserting bytes in front won't work because offsets have already been assembled.
        // The line could be re-assembled by recursively calling DoCPUOpcode, but then
        // the label before the instruction would still be at the odd address.
    }

    return 1;
}


void AsmSHInit(void)
{
    char *p;

    p = AddAsm(versionName, &SH_DoCPUOpcode, NULL, NULL);
    AddCPU(p, "SH1", CPU_SH1, BIG_END, ADDR_24, LIST_24, 8, 0, SH_opcdTab);
    AddCPU(p, "SH2", CPU_SH2, BIG_END, ADDR_24, LIST_24, 8, 0, SH_opcdTab);
    AddCPU(p, "SH3", CPU_SH3, BIG_END, ADDR_24, LIST_24, 8, 0, SH_opcdTab);
    AddCPU(p, "SH4", CPU_SH4, BIG_END, ADDR_24, LIST_24, 8, 0, SH_opcdTab);
}
